# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the policy objects."""

import unittest

import cvxpy as cp
import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.errors import DataError, PortfolioOptimizationError
from cvxportfolio.forecast import HistoricalFactorizedCovariance
from cvxportfolio.tests import CvxportfolioTest

VALUES_IN_TIME_DUMMY_KWARGS = {
    'past_returns': None,
    'current_prices': None,
    'past_volumes': None
}

class TestPolicies(CvxportfolioTest):
    """Test trading policies."""

    def test_hold(self):
        """Test hold policy."""
        hold = cvx.Hold()
        w = pd.Series(0.5, ["AAPL", "CASH"])
        self.assertTrue(np.all(
            hold.values_in_time_recursive(current_weights=w).values
            == w.values))

    def test_rank_and_long_short(self):
        """Test rank-and-long-short policy."""
        w = pd.Series(0.25, ["AAPL", "TSLA", "GOOGL", "CASH"])
        signal = pd.Series([1, 2, 3], ["AAPL", "TSLA", "GOOGL"])
        num_long = 1
        num_short = 1
        target_leverage = 3.0
        rls = cvx.RankAndLongShort(
            signal=signal,
            num_long=num_long,
            num_short=num_short,
            target_leverage=target_leverage,
        )
        wplus = rls.values_in_time_recursive(t=None, current_weights=w)

        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)

        index = pd.date_range("2020-01-01", "2020-01-03")
        signal = pd.DataFrame(
            {
                "AAPL": pd.Series([1, 1.9, 3], index),
                "TSLA": pd.Series([3, 2.1, 1], index),
                "GOOGL": pd.Series([4, 4, 4], index),
            }
        )
        rls = cvx.RankAndLongShort(
            signal=signal,
            num_long=num_long,
            num_short=num_short,
            target_leverage=target_leverage,
        )
        wplus = rls.values_in_time_recursive(t=index[0], current_weights=w)

        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)
        wplus = rls.values_in_time_recursive(t=index[1], current_weights=w)

        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)
        wplus = rls.values_in_time_recursive(t=index[2], current_weights=w)

        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["AAPL"] == 0)
        self.assertTrue(wplus["TSLA"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)

    def test_proportional_trade(self):
        """Test proportional trade policy."""

        a = pd.Series(1., self.returns.columns)
        a.iloc[-1] = 1 - sum(a.iloc[:-1])
        b = pd.Series(-1., self.returns.columns)
        b.iloc[-1] = 1 - sum(b.iloc[:-1])

        targets = pd.DataFrame({self.returns.index[3]: a,
                                self.returns.index[15]: b
                                }).T
        policy = cvx.ProportionalTradeToTargets(targets)

        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        start_portfolio = pd.Series(
            np.random.randn(
                self.returns.shape[1]),
            self.returns.columns)
        start_portfolio.iloc[-1] = 1 - sum(start_portfolio.iloc[:-1])
        for t in self.returns.index[:17]:
            print(t)
            print(start_portfolio)

            wplus = policy.values_in_time_recursive(
                t=t, current_weights=start_portfolio)
            trade = wplus - start_portfolio.values
            start_portfolio = pd.Series(wplus, copy=True)

            if t in targets.index:
                self.assertTrue(np.all(start_portfolio == targets.loc[t]))
        print('trade', trade)
        self.assertTrue(np.allclose(trade, 0.))

    def test_sell_all(self):
        """Test sell-all policy."""
        start_portfolio = pd.Series(
            np.random.randn(
                self.returns.shape[1]),
            self.returns.columns)
        policy = cvx.SellAll()
        t = pd.Timestamp('2022-01-01')
        wplus = policy.values_in_time_recursive(
            t=t, past_returns=self.returns)
        allcash = np.zeros(len(start_portfolio))
        allcash[-1] = 1
        assert isinstance(wplus, pd.Series)
        assert np.allclose(allcash, wplus)

    def test_fixed_trade(self):
        """Test fixed trade policy."""
        fixed_trades = pd.DataFrame(
            np.random.randn(
                len(self.returns),
                self.returns.shape[1]),
            index=self.returns.index,
            columns=self.returns.columns)

        policy = cvx.FixedTrades(fixed_trades)
        t = self.returns.index[123]
        w = pd.Series(0., self.returns.columns)
        wplus = policy.values_in_time_recursive(
            t=t, current_weights=w,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.all(wplus-w == fixed_trades.loc[t]))
        w = wplus-w
        t = pd.Timestamp('1900-01-01')
        wplus = policy.values_in_time_recursive(
            t=t, current_weights=w,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.all(wplus-w == 0.))

    def test_fixed_weights(self):
        """Test fixed weights policy."""
        fixed_weights = pd.DataFrame(
            np.random.randn(
                len(self.returns),
                self.returns.shape[1]),
            index=self.returns.index,
            columns=self.returns.columns)

        policy = cvx.FixedWeights(fixed_weights)
        t = self.returns.index[123]
        wplus = policy.values_in_time_recursive(
            t=t, current_weights=pd.Series(0., self.returns.columns),
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.all(wplus == fixed_weights.loc[t]))

        t = self.returns.index[111]
        wplus = policy.values_in_time_recursive(
            t=t, current_weights=fixed_weights.iloc[110],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertTrue(np.allclose( wplus, fixed_weights.loc[t]))

        t = pd.Timestamp('1900-01-01')
        wplus1 = policy.values_in_time_recursive(t=t, current_weights=wplus,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.all(wplus1 == wplus))

    def test_periodic_rebalance(self):
        """Test periodic rebalance policy."""

        target = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        target /= sum(target)
        rebalancing_times = pd.date_range(
            start=self.returns.index[0], end=self.returns.index[-1], freq='7d')

        policy = cvx.PeriodicRebalance(
            target, rebalancing_times=rebalancing_times)
        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)

        wplus = policy.values_in_time_recursive(
            t=rebalancing_times[0], current_weights=init,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.allclose(wplus, target))

        wplus = policy.values_in_time_recursive(
            t=rebalancing_times[0] + pd.Timedelta('1d'), current_weights=init,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.allclose(wplus, init))

    def test_uniform(self):
        """Test uniform allocation."""
        pol = cvx.Uniform()
        pol.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)

        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)
        wplus = pol.values_in_time_recursive(
            t=self.returns.index[123], current_weights=init,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS,
            )
        self.assertTrue(np.allclose(
            wplus[:-1],
            np.ones(self.returns.shape[1]-1)/(self.returns.shape[1]-1)))

    def test_proportional_rebalance(self):
        """Test the proportional rebalance policy."""

        target = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        target /= sum(target)
        target_matching_times = self.returns.index[::3]

        policy = cvx.ProportionalRebalance(
            target, target_matching_times=target_matching_times)
        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)

        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)

        wplus = policy.values_in_time_recursive(
            t=self.returns.index[1], current_weights=init)
        trade = wplus - init
        wplus2 = policy.values_in_time_recursive(
            t=self.returns.index[2], current_weights=wplus)
        trade2 = wplus2 - wplus
        self.assertTrue(np.allclose(trade, trade2))
        print(trade + trade2 + init, target)
        self.assertTrue(np.allclose(trade + trade2 + init, target))

    def test_adaptive_rebalance(self):
        """Test adaptive rebalance policy."""
        np.random.seed(0)
        target = pd.Series(
            np.random.uniform(
                size=self.returns.shape[1]),
            self.returns.columns)
        target /= sum(target)
        target = pd.DataFrame({ind: target for ind in self.returns.index}).T

        init = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        init /= sum(init)

        for tracking_error in [0.01, .02, .05, .1]:
            policy = cvx.AdaptiveRebalance(
                target, tracking_error=tracking_error)
            wplus = policy.values_in_time_recursive(
                t=self.returns.index[1], current_weights=init)
            self.assertTrue(np.allclose(wplus, target.iloc[0]))

        for tracking_error in [.2, .5]:
            policy = cvx.AdaptiveRebalance(
                target, tracking_error=tracking_error)
            wplus = policy.values_in_time_recursive(
                t=self.returns.index[1], current_weights=init)
            self.assertTrue(np.allclose(wplus - init, 0.))

    def test_single_period_optimization(self):
        """Test basic SPO."""

        return_forecast = cvx.ReturnsForecast()
        risk_forecast = cvx.FullCovariance(
            HistoricalFactorizedCovariance(kelly=False))
        tcost = cvx.TransactionCost(
            a=1E-3/2, b=None, exponent=2)

        policy = cvx.SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
            - tcost,
            constraints=[
                cvx.LongOnly(applies_to_cash=False), cvx.LeverageLimit(1)],
            include_cash_return=False,
            # verbose=True,
            solver=self.default_socp_solver)

        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        # policy.compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        result = policy.values_in_time_recursive(
            t=self.returns.index[121],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:121],
            past_volumes=self.volumes.iloc[:121],
            current_prices=pd.Series(1., self.volumes.columns))

        cvxportfolio_result = pd.Series(result, self.returns.columns)

        print(cvxportfolio_result)

        # print(np.linalg.eigh(self.returns.iloc[:121, :-1].cov().values)[0])

        # REPLICATE WITH CVXPY

        covariance = self.returns.iloc[:121, :-1].cov(ddof=0).values
        w = cp.Variable(self.N)
        cp.Problem(
            cp.Maximize(w[:-1].T @ self.returns.iloc[:121, :-1].mean().values -
            2 * cp.quad_form(w[:-1], covariance) -
            5 * 1E-4 * cp.sum(cp.abs(w - curw)[:-1])),
            [w >= 0, w <= 1, sum(w) == 1]).solve(solver='ECOS')

        cvxpy_result = pd.Series(w.value, self.returns.columns)

        print(cvxpy_result)

        print(cvxportfolio_result - cvxpy_result)
        self.assertTrue(np.allclose(
            cvxportfolio_result - cvxpy_result, 0., atol=1e-5))

    def test_single_period_optimization_solve_twice(self):
        """Test resolve of SPO policy with Cvxpy parameters."""

        return_forecast = cvx.ReturnsForecast()
        risk_forecast = cvx.FullCovariance()

        policy = cvx.SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            - cvx.TransactionCost(a=5 * 1E-4, b=0.),
            constraints=[cvx.LongOnly(), cvx.LeverageLimit(1)],
            # verbose=True,
            solver=self.default_socp_solver)

        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        # policy.compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        result = policy.values_in_time_recursive(
            t=self.returns.index[134],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:134],
            past_volumes=self.volumes.iloc[:134],
            current_prices=pd.Series(1., self.volumes.columns)) - curw

        self.assertFalse(np.allclose(result, 0.))

        curw += result

        result2 = policy.values_in_time_recursive(
            t=self.returns.index[134],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:134],
            past_volumes=self.volumes.iloc[:134],
            current_prices=pd.Series(1., self.volumes.columns)) - curw

        self.assertTrue(np.allclose(result2, 0., atol=1e-7))

    def test_single_period_optimization_infeasible(self):
        """Test SPO policy with infeasible result."""

        return_forecast = cvx.ReturnsForecast()
        risk_forecast = cvx.FullCovariance()
        policy = cvx.SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            - cvx.TransactionCost(a=5 * 1E-4, b=0.),
            constraints=[cvx.LongOnly(), cvx.LeverageLimit(1),
                cvx.MaxWeights(-1)],
            # verbose=True,
            solver=self.default_socp_solver)

        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        # policy.compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        with self.assertRaises(PortfolioOptimizationError):
            policy.values_in_time_recursive(
                t=self.returns.index[134],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:134],
                past_volumes=self.volumes.iloc[:134],
                current_prices=pd.Series(1., self.volumes.columns))

    def test_single_period_optimization_unbounded(self):
        """Test SPO policy with unbounded result."""

        return_forecast = cvx.ReturnsForecast()
        policy = cvx.SinglePeriodOptimization(
            return_forecast,
            constraints=[cvx.LongOnly(applies_to_cash=False)],
            # verbose=True,
            solver=self.default_socp_solver)

        policy.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        # policy.compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        with self.assertRaises(PortfolioOptimizationError):
            policy.values_in_time_recursive(
                t=self.returns.index[134],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:134],
                past_volumes=self.volumes.iloc[:134],
                current_prices=pd.Series(1., self.volumes.columns))

    def test_multi_period_optimization2(self):
        """Test that MPO1 and MPO2 and MPO5 return same if no tcost, and diff.

        if tcost.
        """

        results = []
        for planning_horizon in [1, 2, 5]:
            return_forecast = cvx.ReturnsForecast()
            risk_forecast = cvx.FullCovariance()
            policy = cvx.MultiPeriodOptimization(
                return_forecast
                # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
                - 10 * risk_forecast,
                constraints=[cvx.LongOnly(), cvx.LeverageLimit(1)],
                # verbose=True,
                planning_horizon=planning_horizon,
                solver=self.default_socp_solver)

            policy.initialize_estimator_recursive(
                universe=self.returns.columns,
                trading_calendar=self.returns.index)
            # policy.compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            results.append(policy.values_in_time_recursive(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns).values)
                - curw)

        self.assertTrue(np.allclose(results[0], results[1], atol=1e-4))
        self.assertTrue(np.allclose(results[1], results[2], atol=1e-4))

        # with tcost

        results = []
        for planning_horizon in [1, 2, 5]:
            return_forecast = cvx.ReturnsForecast()
            risk_forecast = cvx.FullCovariance()
            policy = cvx.MultiPeriodOptimization(
                return_forecast
                - 10 * risk_forecast
                # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
                - cvx.TransactionCost(a=25 * 1E-4, b=0.),
                constraints=[cvx.LongOnly(), cvx.LeverageLimit(1)],
                # verbose=True,
                planning_horizon=planning_horizon,
                solver=self.default_socp_solver)

            policy.initialize_estimator_recursive(
                universe=self.returns.columns,
                trading_calendar=self.returns.index)
            # policy.compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            results.append(policy.values_in_time_recursive(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns)) - curw)

        self.assertFalse(np.allclose(results[0], results[1], atol=1e-4))
        self.assertFalse(np.allclose(results[1], results[2], atol=1e-4))

    def test_multi_period_optimization_syntax(self):
        """Test syntax error checks in MultiPeriodOptimization."""
        with self.assertRaises(SyntaxError):
            cvx.MultiPeriodOptimization([cvx.ReturnsForecast()], [])
        with self.assertRaises(SyntaxError):
            cvx.MultiPeriodOptimization([cvx.ReturnsForecast()], [[], []])
        with self.assertRaises(SyntaxError):
            cvx.MultiPeriodOptimization([cvx.ReturnsForecast()], None)
        with self.assertRaises(SyntaxError):
            cvx.MultiPeriodOptimization(cvx.ReturnsForecast())
        cvx.MultiPeriodOptimization(cvx.ReturnsForecast(), planning_horizon=1)

    def test_multi_period_optimization3(self):
        """Check that terminal constraint brings closer to benchmark."""

        np.random.seed(0)
        benchmark = np.random.uniform(size=self.returns.shape[1])
        benchmark /= sum(benchmark)
        benchmark = pd.Series(benchmark, self.returns.columns)

        diff_to_benchmarks = []
        for planning_horizon in [1, 2, 5]:

            return_forecast = cvx.ReturnsForecast()
            risk_forecast = cvx.FullCovariance()
            policy = cvx.MultiPeriodOptimization(
                return_forecast
                - 10 * risk_forecast
                # , power=2)
                - cvx.TransactionCost(a=5 * 1E-4, b=0.),
                constraints=[cvx.LongOnly(), cvx.LeverageLimit(1)],
                # verbose=True,
                terminal_constraint=benchmark,
                planning_horizon=planning_horizon,
                solver=self.default_socp_solver)

            policy.initialize_estimator_recursive(
                universe=self.returns.columns,
                trading_calendar=self.returns.index)
            # policy.compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            diff_to_benchmarks.append(policy.values_in_time_recursive(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns))
                - benchmark)

        self.assertTrue(np.isclose(np.linalg.norm(diff_to_benchmarks[0]), 0.))
        self.assertTrue(np.linalg.norm(
            diff_to_benchmarks[0]) < np.linalg.norm(diff_to_benchmarks[1]))
        self.assertTrue(np.linalg.norm(
            diff_to_benchmarks[1]) < np.linalg.norm(diff_to_benchmarks[2]))

    def test_execute(self):
        """Test the ``execute`` method."""

        policy = cvx.Uniform()
        h = pd.Series(0., self.returns.columns)
        h.iloc[-1] = 10000
        u, t, shares_traded = policy.execute(market_data=self.market_data, h=h)
        print(t, u, shares_traded)
        self.assertTrue(np.isclose(u.sum(), 0.))
        self.assertTrue(t == self.returns.index[-1])
        self.assertTrue(len(set(u.iloc[:-1])) == 1)

        policy = cvx.SinglePeriodOptimization(
            cvx.ReturnsForecast() - 5 * cvx.FullCovariance(),
            [cvx.LongOnly(applies_to_cash=True)],
            solver=self.default_qp_solver)

        market_data = cvx.UserProvidedMarketData(
                    returns=self.returns, volumes=self.volumes,
                    cash_key='cash', base_location=self.datadir,
                    min_history=pd.Timedelta('0d'))

        execution = policy.execute(market_data=market_data, h=h)
        print(execution)
        u, t, shares_traded = execution
        self.assertTrue(shares_traded is None)
        self.assertTrue(np.isclose(u.sum(), 0.))
        self.assertTrue(t == self.returns.index[-1])
        self.assertTrue(u['CSCO'] >= .9)

        h_neg_value = pd.Series(h, copy=True)
        h_neg_value.iloc[-1] = -100
        with self.assertRaises(DataError):
            execution = policy.execute(market_data=market_data, h=h_neg_value)

        h_wrong_uni = pd.Series(h[1:], copy=True)
        with self.assertRaises(ValueError):
            execution = policy.execute(market_data=market_data, h=h_wrong_uni)

        h_wrong_uni = pd.Series(h, copy=True)
        h_wrong_uni['WRONG'] = 1.
        with self.assertRaises(ValueError):
            execution = policy.execute(market_data=market_data, h=h_wrong_uni)

        h_wrong_uni = pd.Series(h, copy=True)
        h_wrong_uni.index = [el + 'SUFFIX' for el in h_wrong_uni.index]
        with self.assertRaises(ValueError):
            execution = policy.execute(market_data=market_data, h=h_wrong_uni)

        # shuffled
        execution = policy.execute(market_data=market_data, h=h)
        h_shuffled = pd.Series(h.iloc[::-1], copy=True)
        execution_shuffled = policy.execute(
            market_data=market_data, h=h_shuffled)
        self.assertTrue(np.all(execution[0] == execution_shuffled[0]))

        # for online
        market_data = cvx.UserProvidedMarketData(
                    returns=self.returns, volumes=self.volumes,
                    cash_key='cash', base_location=self.datadir,
                    min_history=pd.Timedelta('0d'),
                    online_usage=True)

        execution_online = policy.execute(market_data=market_data, h=h)
        self.assertTrue(np.all(execution[0] == execution_online[0]))


if __name__ == '__main__':

    unittest.main(warnings='error') # pragma: no cover



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines hyperparameter objects.

These are used currently as symbolic multipliers of cost terms in Single
and Multi Period Optimization policies and can be iterated (and
optimized over) automatically.
"""

from numbers import Number

import numpy as np
import pandas as pd

# GAMMA_RISK_RANGE = [.5, 1., 2., 5., 10.]
# GAMMA_COST_RANGE = [0., .1, .2, .5, 1., 2., 5., 10.]


__all__ = ['Gamma', 'RangeHyperParameter']


def _resolve_hyperpar(possible_hyperpar):
    """Return current value if input is hyper-parameter, or input itself."""
    if isinstance(possible_hyperpar, HyperParameter):
        return possible_hyperpar.current_value
    return possible_hyperpar

class HyperParameter:
    """Base Hyper Parameter class.

    Implements arithmetic operations between hyper parameters.

    You can sum and multiply HPs between themselves and with scalars,
    and divide by a scalar. Arbitrary algebraic combination of these
    operations are supported.
    """

    def __mul__(self, other):
        if np.isscalar(other) or isinstance(other, HyperParameter) \
                or isinstance(other, pd.Timedelta):
            return CombinedHyperParameter([self], [other])
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        if np.isscalar(other):
            return CombinedHyperParameter([self], [1./other])
        return NotImplemented

    def __truediv__(self, other):
        return self.__div__(other)

    def __add__(self, other):
        if isinstance(other, HyperParameter):
            return CombinedHyperParameter([self, other], [1., 1.])
        return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        return self * (-1)

    def collect_hyperparameters(self):
        """Collect hyper-parameters.

        :returns: Hyperparameters found (self).
        :rtype: list
        """
        return [self]

    @property
    def current_value(self):
        """Current value of the hyper-parameter.

        :returns: Current value.
        :rtype: int, float, pd.Timedelta
        """
        raise NotImplementedError # pragma: no cover

    def __repr__(self):
        return self.__class__.__name__\
            + f'(current_value={self.current_value})'

class CombinedHyperParameter(HyperParameter):
    """Algebraic combination of HyperParameters."""

    def __init__(self, left, right):
        self.left = left
        self.right = right

    @property
    def current_value(self):
        """Current value of the hyper-parameter.

        :returns: Current value.
        :rtype: int or float
        """
        # we unroll the sum() to support non-numeric (Timedeltas, ...)
        summands = list(
            (le.current_value if hasattr(le, 'current_value') else le)
            * (ri.current_value if hasattr(ri, 'current_value') else ri)
            for le, ri in zip(self.left, self.right))
        result = summands[0]
        for el in summands[1:]:
            result += el
        return result

    def collect_hyperparameters(self):
        """Collect (not combined) hyper-parameters.

        :returns: List of found hyper-parameters.
        :rtype: list
        """
        result = []
        for el in self.left + self.right:
            if hasattr(el, 'collect_hyperparameters'):
                result += el.collect_hyperparameters()
        return result

    def __repr__(self):
        """Pretty print.

        .. note::
            This could be improved a lot, it does pretty-printing of algebraic
            expressions in general. It's not perfect but readable.
        """

        # TODO this gives wrong string repr with nested expressions like
        # ``-(cvx.Gamma() * -3 * (cvx.Gamma() - cvx.Gamma()))``
        # internal algebra is correct, though

        def _minus_repr(obj):
            rawrepr = str(obj).lstrip()
            if rawrepr[0] == '-':
                return ' + ' + rawrepr[1:].lstrip()
            if rawrepr[0] == '+':
                return ' - ' + rawrepr[1:].lstrip() # pragma: no cover
            return ' - ' + rawrepr

        def _plus_repr(obj):
            rawrepr = str(obj).lstrip()
            if rawrepr[0] == '-':
                return ' - ' + rawrepr[1:].lstrip()
            if rawrepr[0] == '+':
                return ' + ' + rawrepr[1:].lstrip()
            return ' + ' + rawrepr

        result = ''

        def _with_possible_number(num, other):
            if num == -1.:
                return _minus_repr(other)
            if num == 1.:
                return _plus_repr(other)
            return str(num).rstrip() + ' * ' + str(other).lstrip()

        for left, right in zip(self.left, self.right):
            if isinstance(left, Number):
                result += _with_possible_number(left, right) # pragma: no cover
            else:
                result += _with_possible_number(right, left)

        return result.strip()

class RangeHyperParameter(HyperParameter):
    """Range Hyper Parameter.

    This is not meant to be used directly, look at its subclasses for
    ones that you can use.
    """

    def __init__(self, values_range, current_value):
        if not current_value in values_range:
            raise SyntaxError('Initial value must be in the provided range')
        self.values_range = list(values_range)
        self._index = self.values_range.index(current_value)

    @property
    def current_value(self):
        """Current value of the hyper-parameter.

        :returns: Current value.
        :rtype: int or float
        """
        return self.values_range[self._index]

    def _increment(self):
        if self._index == len(self.values_range) - 1:
            raise IndexError
        self._index += 1

    def _decrement(self):
        if self._index == 0:
            raise IndexError
        self._index -= 1


class Gamma(HyperParameter):
    """Generic multiplier."""

    def __init__(self, initial_value = 1., increment = 1.1):
        self._initial_value = initial_value
        self._spacing = increment
        self._index = 0

    @property
    def current_value(self):
        """Current value of the hyper-parameter.

        :returns: Current value.
        :rtype: int or float
        """
        return self._initial_value * (self._spacing ** self._index)

    def _increment(self):
        # if self._index == len(self.values_range) - 1:
        #     raise IndexError
        self._index += 1

    def _decrement(self):
        # if self._index == 0:
        #     raise IndexError
        self._index -= 1
#
# class GammaRisk(RangeHyperParameter):
#     """Multiplier of a risk term."""
#
#     def __init__(self, values_range=GAMMA_RISK_RANGE, current_value=1.):
#         super().__init__(values_range, current_value)
#
#
# class GammaTrade(RangeHyperParameter):
#     """Multiplier of a transaction cost term."""
#
#     def __init__(self, values_range=GAMMA_COST_RANGE, current_value=1.):
#         super().__init__(values_range, current_value)
#
#
# class GammaHold(RangeHyperParameter):
#     """Multiplier of a holding cost term."""
#
#     def __init__(self, values_range=GAMMA_COST_RANGE, current_value=1.):
#         super().__init__(values_range, current_value)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains trading policies that can be back-tested."""

import copy
import logging
import warnings

import cvxpy as cp
import numpy as np
import pandas as pd

from .errors import (ConvexityError, ConvexSpecificationError, DataError,
                     MissingTimesError, PortfolioOptimizationError)
from .estimator import DataEstimator, Estimator
from .forecast import HistoricalMeanVolume
from .returns import CashReturn
from .utils import flatten_heterogeneous_list

__all__ = [
    "AllCash",
    "Hold",
    "SellAll",
    "FixedTrades",
    "PeriodicRebalance",
    "ProportionalRebalance",
    "AdaptiveRebalance",
    "SinglePeriodOptimization",
    "MultiPeriodOptimization",
    "MarketBenchmark",
    "ProportionalTradeToTargets",
    "RankAndLongShort",
    "FixedWeights",
    "Uniform",
    "SinglePeriodOpt",
    "MultiPeriodOpt",
]

logger = logging.getLogger(__name__)

class Policy(Estimator):
    """Base trading policy class, defines execute method."""

    def execute(self, h, market_data, t=None):
        """Execute trading policy at current or user-specified time.

        Return the (*e.g.*, dollar) trade vector :math:`u`, the timestamp
        of execution (for double check in case you don't pass it), and a Pandas
        Series of the number of shares to trade, if you pass a Market
        Data server which provides open prices (or None).

        :param h: Holdings vector, in dollars, including the cash account
            (the last element).
        :type h: pandas.Series
        :param market_data: :class:`MarketData` instance used to provide
            data to the policy
        :type market_data: cvxportfolio.MarketData instance
        :param t: Time at which we execute. If None (the default), the
            last timestamp in the trading calendar provided by the
            :class:`MarketData` instance is used. Note: if you use a default
            market data server, you probably want to set their ``online_usage``
            argument to ``True``.
        :type t: pandas.Timestamp or None

        :raises cvxportfolio.errors.DataError: Holdings vector sum to a
            negative value or don't match the market data server's universe.

        :returns: u, t, shares_traded
        :rtype: pandas.Series, pandas.Timestamp, pandas.Series
        """

        trading_calendar = market_data.trading_calendar()

        if t is None:
            t = trading_calendar[-1]

        assert t in trading_calendar

        v = np.sum(h)

        if v < 0.:
            raise DataError(
                f"Holdings provided to {self.__class__.__name__}.execute "
                + " have negative sum.")

        past_returns, _, past_volumes, _, current_prices = market_data.serve(t)

        if sorted(h.index) != sorted(past_returns.columns):
            raise DataError(
                "Holdings provided don't match the universe"
                " implied by the market data server.")

        h = h[past_returns.columns]
        w = h / v

        # consider adding caching logic here
        self.initialize_estimator_recursive(
            universe=past_returns.columns,
            trading_calendar=trading_calendar[trading_calendar >= t])

        w_plus = self.values_in_time_recursive(
            t=t, past_returns=past_returns, past_volumes=past_volumes,
            current_weights=w, current_portfolio_value=v,
            current_prices=current_prices)

        # this could be optional, currently unused)
        self.finalize_estimator_recursive()

        z = w_plus - w
        u = z * v

        if current_prices is not None:
            shares_traded =  pd.Series(np.round(u.iloc[:-1] / current_prices),
                dtype=int)
        else:
            shares_traded = None

        return u, t, shares_traded


class Hold(Policy):
    """Hold initial portfolio, don't trade."""

    def values_in_time( # pylint: disable=arguments-differ
            self, current_weights, **kwargs):
        """Return current_weights.

        :param current_weights: Current weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Same current weights.
        :rtype: pandas.Series
        """
        return current_weights

class AllCash(Policy):
    """Allocate all weight to cash.

    This is the default benchmark used in :class:`SinglePeriodOptimization` and
    :class:`MultiPeriodOptimization` policies.
    """

    def values_in_time( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Return all cash weights.

        :param past_returns: Past market returns (used to infer universe).
        :type past_returns: pandas.DataFrame
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: All cash weights.
        :rtype: pandas.Series
        """
        result = pd.Series(0., past_returns.columns)
        result.iloc[-1] = 1.
        return result

class MarketBenchmark(Policy):
    """Allocation weighted by last year's average market traded volumes.

    This policy provides an approximation of a market capitalization-weighted
    allocation, by using the average of traded volumes in units of value (e.g.,
    USDOLLAR) over the previous year as proxy.

    .. versionadded:: 1.2.0

        We added the ``mean_volume_forecast`` parameter.

    :param mean_volume_forecast: Forecaster class that computes the average
        of historical volumes. You can also pass a DataFrame containing
        your own forecasts computed externally. Default is
        :class:`cvxportfolio.forecast.HistoricalMeanVolume` which is
        instantiated with parameter ``rolling=pd.Timedelta('365.24d')``
        (that's one solar year in number of days). If you
        want to provide a different forecaster, or change the parameters (like
        adding exponential smoothing) you should instantiate the forecaster
        class and pass the instance.
    :type mean_volume_forecast: pandas.DataFrame, cvx.forecast.BaseForecast
        class or instance
    """

    def __init__(self, mean_volume_forecast=HistoricalMeanVolume):
        if isinstance(mean_volume_forecast, type):
            mean_volume_forecast = mean_volume_forecast(
                rolling=pd.Timedelta('365.24d'))
        self.mean_volume_forecast = DataEstimator(
            mean_volume_forecast, data_includes_cash=False)

    def values_in_time( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Return market benchmark weights.

        :param past_returns: Past market returns (used to infer universe with
            cash).
        :type past_returns: pandas.DataFrame
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :raises cvxportfolio.errors.DataError: Market data does not include
            market volumes.

        :returns: Market benchmark weights.
        :rtype: pandas.Series
        """

        meanvolumes = self.mean_volume_forecast.current_value
        result = np.zeros(len(meanvolumes) + 1)
        result[:-1] = meanvolumes / sum(meanvolumes)
        return pd.Series(result, index=past_returns.columns)

class RankAndLongShort(Policy):
    """Rank assets by signal; long highest and short lowest.

    :param signal: time-indexed DataFrame of signal for all symbols
        excluding cash. At each point in time the num_long assets with
        highest signal will have equal positive weight, and the
        num_short assets with lower signal will have equal negative
        weight. If two or more assets have the same signal value and
        they are on the boundary of either the top or bottom set,
        alphanumerical ranking will prevail.
    :type signal: pd.DataFrame
    :param num_long: number of assets to long, default 1; if specified
        as Series it must be indexed by time.
    :type num_long: int or pd.Series
    :param num_short: Number of assets to short, default 1; if specified
        as Series it must be indexed by time.
    :type num_short: int or pd.Series)
    :param target_leverage: leverage of the resulting portfolio, default
        1; if specified as Series it must be indexed by time.
    :type target_leverage: float or pd.Series
    """

    def __init__(self, signal, num_long=1, num_short=1, target_leverage=1.):
        """Define sub-estimators at class attribute level."""
        self.num_long = DataEstimator(num_long)
        self.num_short = DataEstimator(num_short)
        self.signal = DataEstimator(signal)
        self.target_leverage = DataEstimator(target_leverage)

    def values_in_time( # pylint: disable=arguments-differ
            self, current_weights, **kwargs):
        """Get allocation weights.

        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Rank and long-short weights.
        :rtype: pandas.Series
        """

        sorted_ret = pd.Series(
            self.signal.current_value, current_weights.index[:-1]
        ).sort_values()
        short_positions = sorted_ret.index[: self.num_short.current_value]
        long_positions = sorted_ret.index[-self.num_long.current_value:]

        target_weights = pd.Series(0., index=current_weights.index)
        target_weights[short_positions] = -1.
        target_weights[long_positions] = 1.

        target_weights /= sum(abs(target_weights))
        target_weights *= self.target_leverage.current_value

        # cash is always 1.
        target_weights[current_weights.index[-1]] = 1.

        return target_weights


class ProportionalTradeToTargets(Policy):
    """Trade in equal proportion to match target weights in time.

    Initially, it loads the list of trading days and so at each day it
    knows how many are missing before the next target's day, and trades
    in equal proportions to reach those targets. If there are no targets
    remaining it defaults to not trading.

    :param targets: time-indexed DataFrame of target weight vectors at
        given points in time (e.g., start of each month).
    :type targets: pandas.DataFrame
    """

    def __init__(self, targets):
        self.targets = targets
        self.trading_days = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, trading_calendar, **kwargs):
        """Initialize policy instance with updated trading_calendar.

        :param trading_calendar: Future (including current) trading calendar.
        :type trading_calendar: pandas.DatetimeIndex
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self.trading_days = trading_calendar

    def values_in_time( # pylint: disable=arguments-differ
            self, t, current_weights, **kwargs):
        """Get current allocation weights.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :raises ValueError: Some target weights provided do not sum to 1.

        :returns: Allocation weights.
        :rtype: pandas.Series
        """

        next_targets = self.targets.loc[self.targets.index >= t]
        if not np.allclose(next_targets.sum(1), 1.):
            raise ValueError(
                "The target weights provided to %s at time %s"
                + " do not sum to 1.", self.__class__.__name__, t)
        if len(next_targets) == 0:
            return current_weights
        next_target = next_targets.iloc[0]
        next_target_day = next_targets.index[0]
        trading_days_to_target = len(self.trading_days[(
            self.trading_days >= t) & (self.trading_days < next_target_day)])
        if trading_days_to_target == 0:
            return current_weights
        return current_weights + (
            next_target - current_weights) / trading_days_to_target


class SellAll(AllCash):
    """Sell all assets to cash.

    Alias of :class:`AllCash`.
    """


class FixedTrades(Policy):
    """Each day trade the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    :param trades_weights: target trade weights :math:`z_t` to trade at each
        period. If constant in time use a pandas Series indexed by the assets'
        names, including the cash account name (``cash_key`` option to
        :class:`MarketSimulator`). If varying in time, use a pandas DataFrame
        with datetime index and as columns the assets names including cash.
        If a certain time in the backtest is not present in the data provided
        the policy defaults to not trading in that period.
    :type trades_weights: pd.Series or pd.DataFrame
    """

    def __init__(self, trades_weights):
        """Trade fixed trade weigths in time."""
        self.trades_weights = DataEstimator(
            trades_weights, data_includes_cash=True)

    def values_in_time_recursive( # pylint: disable=arguments-differ
            self, t, current_weights, **kwargs):
        """Get current allocation weights.

        We redefine the recursive version of :meth:`values_in_time` because
        we catch an exception thrown by a sub-estimator.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Allocation weights.
        :rtype: pandas.Series
        """
        try:
            super().values_in_time_recursive(
                t=t, current_weights=current_weights, **kwargs)
            result = current_weights + pd.Series(
                self.trades_weights.current_value, current_weights.index)
        except MissingTimesError:
            logger.info("%s didn't trade at time %s because it couldn't find"
                + " trade weights among the provided ones.",
                self.__class__.__name__, t)
            result = current_weights
        self._current_value = result
        return result


class FixedWeights(Policy):
    """Each day trade to the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    :param target_weights: target weights :math:`w_t^+` to trade to at each
        period. If constant in time use a pandas Series indexed by the assets'
        names, including the cash account name (``cash_key`` option
        to the simulator). If varying in time, use a pandas DataFrame
        with datetime index and as columns the assets names including cash.
        If a certain time in the backtest is not present in the data provided
        the policy defaults to not trading in that period.
    :type target_weights: pd.Series or pd.DataFrame
    """

    def __init__(self, target_weights):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        self.target_weights = DataEstimator(
            target_weights, data_includes_cash=True)

    def values_in_time_recursive( # pylint: disable=arguments-differ
            self, t, current_weights, **kwargs):
        """Get current allocation weights.

        We redefine the recursive version of :meth:`values_in_time` because
        we catch an exception thrown by a sub-estimator.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time_recursive`.
        :type kwargs: dict

        :returns: Allocation weights.
        :rtype: pandas.Series
        """
        try:
            super().values_in_time_recursive(
                t=t, current_weights=current_weights, **kwargs)
            result = current_weights + pd.Series(
                self.target_weights.current_value, current_weights.index
                ) - current_weights
        except MissingTimesError:
            logger.info("%s didn't trade at time %s because it couldn't find"
                + " target weights among the provided ones.",
                self.__class__.__name__, t)
            result = current_weights
        self._current_value = result
        return result


class Uniform(FixedWeights):
    """Uniform allocation on non-cash assets.

    :param leverage: Leverage of the allocation.
    :type leverage: float
    """

    # pylint: disable=super-init-not-called
    def __init__(self, leverage=1.):
        self.leverage = leverage
        # then we re-define the target weights for each universe in the method
        # below, so we provide the same interface to the parent class'
        # values_in_time_recursive
        self.target_weights = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize this estimator.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        target_weights = pd.Series(1., universe)
        target_weights.iloc[-1] = 0
        target_weights /= sum(target_weights)
        target_weights *= self.leverage
        target_weights.iloc[-1] = 1. - target_weights.sum()
        self.target_weights = DataEstimator(target_weights)


class PeriodicRebalance(FixedWeights):
    """Track a target weight vector rebalancing at given times.

    This calls `FixedWeights`. If you want to change the target in time
    use that policy directly.

    :param target: Allocation weights to rebalance to.
    :type target: pandas.Series
    :param rebalancing_times: Times at which we rebalance.
    :type rebalancing_times: pandas.DateTimeIndex):
    """

    def __init__(self, target, rebalancing_times):
        target_weights = pd.DataFrame(
            {el: target for el in rebalancing_times}).T
        super().__init__(target_weights)


class ProportionalRebalance(ProportionalTradeToTargets):
    """Trade proportionally in time to track fixed target weights at times.

    This calls `ProportionalTradeToTargets`. If you want to change
    the target in time use that policy directly.

    :param target: Allocation weights to rebalance to.
    :type target: pandas.Series
    :param rebalancing_times: Times at which we rebalance.
    :type rebalancing_times: pandas.DateTimeIndex):
    """

    def __init__(self, target, target_matching_times):
        targets = pd.DataFrame({el: target for el in target_matching_times}).T
        super().__init__(targets)


class AdaptiveRebalance(Policy):
    """Rebalance portfolio when deviates too far from target.

    We use the 2-norm as trigger for rebalance. You may want to
    calibrate ``tracking_error`` for your application
    by backtesting this policy, *e.g.*, to get your desired turnover.

    :param target: target weights to rebalance to.
        It is assumed a constant if it is a Series. If it varies in
        time (you must specify it for every trading day) pass a
        DataFrame indexed by time.
    :type target: pd.Series or pd.DataFrame
    :param tracking_error: we trade to match the target
        weights whenever the 2-norm of our weights minus the
        target is larger than this. Pass a Series if you want to vary it in
        time.
    :type tracking_error: pd.Series or pd.DataFrame
    """

    def __init__(self, target, tracking_error):
        self.target = DataEstimator(target)
        self.tracking_error = DataEstimator(tracking_error)

    def values_in_time( # pylint: disable=arguments-differ
            self, current_weights, **kwargs):
        """Get target allocation weights.

        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Allocation weights.
        :rtype: pandas.Series
        """
        if np.linalg.norm(current_weights - self.target.current_value) >\
                self.tracking_error.current_value:
            return self.target.current_value
        return current_weights

# pylint: disable=too-many-instance-attributes
class MultiPeriodOptimization(Policy):
    r"""Multi Period Optimization policy.

    Implements the model developed in :paper:`chapter 5, page 49 <section.5.2>`
    of the paper.
    You specify the objective terms
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify lists
    of constraints. There are two ways to do it. You either
    define the same objective terms and costraints for each
    step of the multi-period problem, or you define a different
    objective term and different list of constraints for each step.
    In addition we offer a `terminal_constraint` argument to
    simply impose that at the last step in the optimization the
    post-trade weights match the given weights (see page 51).

    When it computes the trajectory of weights for the future
    it only returns the first step (to the Simulator, typically).
    The future steps (planning horizon) are by default not returned.

    :param objective: These will be maximized;
        if you pass a single expression of Cost it is understood as the
        same for all steps; if it's a list you must also pass a list of lists
        for `constraints`, each term represents the cost for each step of the
        optimization (starting from the first, i.e., today) and the length of
        the list is used as planning_horizon (the value you pass there will be
        ignored)
    :type objective: algebra of Cost or list of
    :param constraints: These will be
        imposed on the optimization. Default []. Pass this as a list of
        lists of the same length as `objective` to specify different
        constraints at different time steps.
    :type constraints: list of Constraints or list of those
    :param planning_horizon:  How many steps in the future we
        plan for. Ignored if passing `objective` and `constraints` as lists.
        Default is None.
    :type planning_horizon: int or None
    :param terminal_constraint: If you pass a Series to this
        (default is None) it will impose that at the last step of the multi
        period optimization the post-trade weights are equal to this.
    :type terminal_constraint: pd.Series or None
    :param include_cash_return: Whether to automatically include the
        ``CashReturn`` term in the objective, with default parameters.
        Default is ``True``.
    :type include_cash_return: bool
    :param benchmark: Benchmark weights to use in the risk model and
        other terms that need it. You can use any policy here. Suggested ones
        ones are ``AllCash``, the default, ``Uniform``
        (uniform allocation on non-cash assets),
        and ``MarketBenchmark``, which approximates the market-weighted
        portfolio.
    :type benchmark: :class:`Policy` class or instance
    :param kwargs: Any extra argument will be passed to cvxpy.Problem.solve,
        so you can choose a solver and pass parameters to it.
    :type kwargs: dict

    :raises SyntaxError: If the format of provided objective and constraints
        is not right.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self, objective, constraints=(), include_cash_return=True,
        planning_horizon=None, terminal_constraint=None,
        benchmark=AllCash, **kwargs):
        if hasattr(objective, '__iter__'):
            if not (hasattr(constraints, '__iter__') and len(constraints
                    ) and (hasattr(constraints[0], '__iter__') and len(
                    objective) == len(constraints))):
                raise SyntaxError(
                    'If you pass the objective as a list, constraints should'
                    + ' be a list of lists of the same length.')
            self._planning_horizon = len(objective)
            self.objective = objective
            self.constraints = constraints
        else:
            if not np.isscalar(planning_horizon):
                raise SyntaxError(
                    'If `objective` and `constraints` are the same for '
                    + 'all steps you must specify `planning_horizon`.')
            self._planning_horizon = planning_horizon
            self.objective = [objective._copy_keeping_multipliers()
                if hasattr(objective, '_copy_keeping_multipliers')
                    else copy.deepcopy(objective) for i in range(
                    planning_horizon)] if planning_horizon > 1 else [objective]
            self.constraints = [copy.deepcopy(constraints) for i in range(
                planning_horizon)] if planning_horizon > 1 else [constraints]

        self._include_cash_return = include_cash_return
        if self._include_cash_return:
            self.objective = [el + CashReturn() for el in self.objective]
        self.terminal_constraint = terminal_constraint

        if isinstance(benchmark, (pd.DataFrame, pd.Series)):
            self.benchmark = DataEstimator(benchmark, data_includes_cash=True)
        else:
            self.benchmark = benchmark() if isinstance(benchmark, type
                ) else benchmark

        self.cvxpy_kwargs = kwargs

        # redefined below
        self._cvxpy_objective = 0
        self._cvxpy_constraints = []
        self._problem = None
        self._w_bm = None
        self._w_current = None
        self._z_at_lags = None
        self._w_plus_at_lags = None
        self._w_plus_minus_w_bm_at_lags = None
        self._cache = {}

    def _compile_to_cvxpy(self):
        """Compile all cvxpy expressions and the problem."""
        self._cvxpy_objective = [
            el.compile_to_cvxpy(
                self._w_plus_at_lags[i], self._z_at_lags[i],
                self._w_plus_minus_w_bm_at_lags[i])
            for i, el in enumerate(self.objective)]
        for el, term in zip(self.objective, self._cvxpy_objective):
            if not term.is_dcp():
                raise ConvexSpecificationError(el)
            if not term.is_concave():
                raise ConvexityError(el)
        self._cvxpy_objective = sum(self._cvxpy_objective)

        def _compile_and_check_constraint(constr, i):
            result = constr.compile_to_cvxpy(
                self._w_plus_at_lags[i], self._z_at_lags[i],
                self._w_plus_minus_w_bm_at_lags[i])
            for el in (result if hasattr(result, '__iter__') else [result]):
                if not el.is_dcp():
                    raise ConvexSpecificationError(constr)
            return result

        self._cvxpy_constraints = [
            flatten_heterogeneous_list([
                _compile_and_check_constraint(constr, i) for constr in el])
            for i, el in enumerate(self.constraints)]

        self._cvxpy_constraints = sum(self._cvxpy_constraints, [])
        self._cvxpy_constraints += [cp.sum(z) == 0 for z in self._z_at_lags]
        w = self._w_current
        for i in range(self._planning_horizon):
            self._cvxpy_constraints.append(
                self._w_plus_at_lags[i] == self._z_at_lags[i] + w)
            self._cvxpy_constraints.append(
                self._w_plus_at_lags[i] - self._w_bm == \
                    self._w_plus_minus_w_bm_at_lags[i])
            w = self._w_plus_at_lags[i]
        if not self.terminal_constraint is None:
            self._cvxpy_constraints.append(w == self.terminal_constraint)
        self._problem = cp.Problem(cp.Maximize(
            self._cvxpy_objective), self._cvxpy_constraints)
        if not self._problem.is_dcp():  # dpp=True)
            raise SyntaxError(
                "The optimization problem compiled by %s"
                + " does not follow the convex optimization rules."
                + " This should not happen if you're using the default "
                + " cvxportfolio terms and is probably due to a"
                + " mis-specified custom term.", self.__class__.__name__)

    def initialize_estimator_recursive( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize the policy object with the trading universe.

        We redefine the recursive version of :meth:`initialize_estimator`
        because we initialize all objective and constraint objects.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """

        for obj in self.objective:
            obj.initialize_estimator_recursive(universe=universe, **kwargs)
        for constr_at_lag in self.constraints:
            for constr in constr_at_lag:
                constr.initialize_estimator_recursive(
                    universe=universe, **kwargs)

        self.benchmark.initialize_estimator_recursive(
            universe=universe, **kwargs)

        self._w_bm = cp.Parameter(len(universe))

        self._w_current = cp.Parameter(len(universe))
        self._z_at_lags = [cp.Variable(len(universe))
                          for i in range(self._planning_horizon)]
        self._w_plus_at_lags = [cp.Variable(
            len(universe)) for i in range(self._planning_horizon)]
        self._w_plus_minus_w_bm_at_lags = [cp.Variable(
            len(universe)) for i in range(self._planning_horizon)]

        # simulator will overwrite this with cache loaded from disk
        self._cache = {}

        self._compile_to_cvxpy()

    def finalize_estimator_recursive(self, **kwargs):
        """Finalize all objects in this policy's estimator tree.

        :param kwargs: Arguments.
        :type kwargs: dict
        """
        for obj in self.objective:
            obj.finalize_estimator_recursive(**kwargs)
        for constr_at_lag in self.constraints:
            for constr in constr_at_lag:
                constr.finalize_estimator_recursive(**kwargs)
        self.benchmark.finalize_estimator_recursive(**kwargs)

    def values_in_time_recursive( # pylint: disable=arguments-differ
            self, t, current_weights, current_portfolio_value, **kwargs):
        """Update all cvxpy parameters, solve, and return allocation weights.

        We redefine the recursive version of :meth:`values_in_time`
        because we evaluate all objective and constraint objects.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param current_portfolio_value: Current total value of the portfolio.
        :type current_portfolio_value: float
        :param kwargs: Other arguments to :meth:`values_in_time_recursive`.
        :type kwargs: dict

        :raises cvxportfolio.errors.DataError: Current portfolio value is
            negative.
        :raises cvxportfolio.errors.PortfolioOptimizationError: The portfolio
            optimization failed: either the problem was infeasible on
            unbounded, or a numerical error occurred. In the latter case the
            original Cvxpy exception is re-raised.

        :returns: Allocation weights.
        :rtype: pandas.Series
        """

        if not current_portfolio_value > 0:
            raise DataError(
                f"Policy {self.__class__.__name__} was evaluated at "
                + "{t} with negative portfolio value.")
        assert np.isclose(sum(current_weights), 1)

        for i, obj in enumerate(self.objective):
            obj.values_in_time_recursive(
                t=t, current_weights=current_weights,
                current_portfolio_value=current_portfolio_value,
                mpo_step=i, cache=self._cache,
                **kwargs)

        for i, constr_at_lag in enumerate(self.constraints):
            for constr in constr_at_lag:
                constr.values_in_time_recursive(
                    t=t, current_weights=current_weights,
                    current_portfolio_value=current_portfolio_value,
                    mpo_step=i, cache=self._cache, **kwargs)

        self.benchmark.values_in_time_recursive(
            t=t, current_weights=current_weights,
            current_portfolio_value=current_portfolio_value,
            **kwargs)

        self._w_bm.value = np.array(self.benchmark.current_value.values)\
             if hasattr(self.benchmark.current_value, 'values'
            ) else np.array(self.benchmark.current_value)
        self._w_current.value = current_weights.values

        try:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message='Solution may be inaccurate')
                # suppress cvxpy 1.4 ECOS deprecation warnings
                if cp.__version__[:3] == '1.4':
                    warnings.filterwarnings("ignore", category=FutureWarning)
                self._problem.solve(**self.cvxpy_kwargs)
        except cp.SolverError as exc:
            raise PortfolioOptimizationError(
                "Numerical solver for policy %s at time %s failed;"
                + " try changing it, relaxing some constraints,"
                + " or removing costs.", self.__class__.__name__, t) from exc

        if self._problem.status in ["unbounded", "unbounded_inaccurate"]:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time "
                + f"{t} resulted in an unbounded problem.")

        if self._problem.status in ["infeasible", 'infeasible_inaccurate']:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time "
                + f"{t} resulted in an infeasible problem.")

        result = current_weights + pd.Series(
            self._z_at_lags[0].value, current_weights.index)
        self._current_value = result
        return result

    def collect_hyperparameters(self):
        """Collect hyper-parameters in the policy definition.

        :returns: List of :class:`cvxportfolio.hyperparameters.HyperParameter`
            instances.
        :rtype: list
        """
        result = []
        for el in self.objective:
            result += el.collect_hyperparameters()
        for el in self.constraints:
            for constr in el:
                result += constr.collect_hyperparameters()
        return result


class SinglePeriodOptimization(MultiPeriodOptimization):
    r"""Single Period Optimization policy.

    Implements the model developed in :paper:`chapter 4, page 43 <section.4.8>`
    of the paper.
    You specify the objective terms
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify lists
    of constraints.

    :param objective: This algebraic combination of cvxportfolio cost objects
        will be maximized
    :type objective: CombinedCost
    :param constraints: These will be imposed on the optimization. Default [].
    :type constraints: list of Constraints
    :param include_cash_return: Whether to automatically include the
        ``CashReturn`` term in the objective, with default parameters.
        Default is ``True``.
    :type include_cash_return: bool
    :param benchmark: Benchmark weights to use in the risk model and
        other terms that need it. You can use any policy here. Suggested ones
        ones are ``AllCash``, the default, ``Uniform``
        (uniform allocation on non-cash assets),
        and ``MarketBenchmark``, which approximates the market-weighted
        portfolio.
    :type benchmark: :class:`Policy` class or instance
    :param kwargs: Any extra argument will be passed to cvxpy.Problem.solve,
        so you can choose a solver and pass parameters to it.
    :type kwargs: dict
    """

    def __init__(self, objective, constraints=(),
                include_cash_return=True, benchmark=AllCash, **kwargs):
        super().__init__(
            [objective], [constraints],
            include_cash_return=include_cash_return,
            benchmark=benchmark, **kwargs)

    def __repr__(self):
        return self.__class__.__name__ + '(' \
            + 'objective=' + str(self.objective[0]) \
            + ', constraints=' + str(self.constraints[0])\
            + ', benchmark=' + str(self.benchmark)\
            + ', cvxpy_kwargs=' + str(self.cvxpy_kwargs)\
            + ')'

# Aliases


class SinglePeriodOpt(SinglePeriodOptimization):
    """Alias of :class:`SinglePeriodOptimization`.

    As it was defined originally in :paper:`section 6.1 <section.6.1>` of the
    paper.
    """

class MultiPeriodOpt(MultiPeriodOptimization):
    """Alias of :class:`MultiPeriodOptimization`.

    As it was defined originally in :paper:`section 6.1 <section.6.1>` of the
    paper.
    """



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the constraints objects."""

import unittest

import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.tests import CvxportfolioTest

VALUES_IN_TIME_DUMMY_KWARGS = {
    'current_weights': None,
    'past_returns': None,
    'current_prices': None,
    'past_volumes': None
}

class TestConstraints(CvxportfolioTest):
    """Test Cvxportfolio constraint objects."""

    def _build_constraint(self, constraint, t=None):
        """Initialize constraint, build expression, and point it to time."""
        constraint.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        cvxpy_expression = constraint.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        constraint.values_in_time_recursive(
            t=pd.Timestamp("2020-01-01") if t is None else t,
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS
            )
        return cvxpy_expression

    def test_long_only(self):
        """Test long-only constraint."""
        model = cvx.LongOnly()
        cons = self._build_constraint(model)
        self.w_plus.value = np.ones(self.N)
        self.assertTrue(cons.value())
        self.w_plus.value = -np.ones(self.N)
        self.assertFalse(cons.value())

    def test_nocash(self):
        """Test no cash constraint."""
        model = cvx.NoCash()
        cons = self._build_constraint(model)
        self.w_plus.value = np.ones(self.N)
        self.assertFalse(cons.value())
        self.w_plus.value = -np.ones(self.N)
        self.assertFalse(cons.value())
        self.w_plus.value[-1] = 0.
        self.assertTrue(cons.value())

    def test_long_cash(self):
        """Test long-cash constraint."""
        model = cvx.LongCash()
        cons = self._build_constraint(model)
        self.w_plus.value = np.ones(self.N)
        self.assertTrue(cons.value())
        tmp = np.ones(self.N)
        tmp[-1] = -1
        self.w_plus.value = tmp
        self.assertTrue(cons.is_dcp())
        self.assertFalse(cons.value())

    def test_min_cash(self):
        """Test min-cash constraint."""
        model = cvx.MinCashBalance(10000)  # USD
        cons = self._build_constraint(model)
        self.w_plus.value = np.zeros(self.N)
        self.w_plus.value[-1] = 1
        model.values_in_time_recursive(t=pd.Timestamp(
            "2020-01-01"), current_portfolio_value=10001,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=pd.Timestamp(
            "2020-01-01"), current_portfolio_value=9999,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_dollar_neutral(self):
        """Test dollar-neutral constraint."""
        model = cvx.DollarNeutral()
        cons = self._build_constraint(model)
        tmpvalue = np.zeros(self.N)
        tmpvalue[-1] = 1 - sum(tmpvalue[:-1])
        self.w_plus.value = tmpvalue
        self.assertTrue(cons.value())
        tmpvalue = np.ones(self.N)
        tmpvalue[-1] = 1 - sum(tmpvalue[:-1])
        self.w_plus.value = tmpvalue
        self.assertFalse(cons.value())

    def test_leverage_limit(self):
        """Test leverage limit constraint."""
        model = cvx.LeverageLimit(2)
        cons = self._build_constraint(model)
        self.w_plus.value = np.ones(self.N) / self.N
        self.assertTrue(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())
        model = cvx.LeverageLimit(7)
        cons = self._build_constraint(model)
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_leverage_limit_in_time(self):
        """Test leverage limit constraint with time-varying limit."""
        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = 7
        model = cvx.LeverageLimit(limits)
        cons = self._build_constraint(model, t=self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=self.returns.index[2],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_max_weights(self):
        """Test max weights constraint."""
        model = cvx.MaxWeights(2)
        cons = self._build_constraint(model)
        self.w_plus.value = np.ones(self.N) / self.N
        self.assertTrue(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())

        model = cvx.MaxWeights(7)
        cons = self._build_constraint(model)

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = 7

        model = cvx.MaxWeights(limits)
        cons = self._build_constraint(model, t=self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=self.returns.index[2],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_min_weights(self):
        """Test min weights constraint."""
        model = cvx.MinWeights(2)
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())
        model = cvx.MinWeights(-3)
        cons = self._build_constraint(model, self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = -3
        model = cvx.MinWeights(limits)
        cons = self._build_constraint(model, t=self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=self.returns.index[2],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_max_bm_dev(self):
        """Test max benchmark deviation constraint."""
        model = cvx.MaxBenchmarkDeviation(2)
        cons = self._build_constraint(model)
        self.w_plus_minus_w_bm.value = np.ones(self.N) / self.N
        self.assertTrue(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertFalse(cons.value())

        model = cvx.MaxBenchmarkDeviation(7)
        cons = self._build_constraint(model)

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = 7

        model = cvx.MaxBenchmarkDeviation(limits)
        cons = self._build_constraint(model, t=self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=self.returns.index[2],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_min_bm_dev(self):
        """Test min benchmark deviation constraint."""
        model = cvx.MinBenchmarkDeviation(2)
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus_minus_w_bm.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertFalse(cons.value())
        model = cvx.MinBenchmarkDeviation(-3)
        cons = self._build_constraint(model, self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = -3
        model = cvx.MinBenchmarkDeviation(limits)
        cons = self._build_constraint(model, t=self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus_minus_w_bm.value = tmp
        self.assertTrue(cons.value())
        model.values_in_time_recursive(t=self.returns.index[2],
            current_portfolio_value=1000,
            **VALUES_IN_TIME_DUMMY_KWARGS)
        self.assertFalse(cons.value())

    def test_factor_max_limit(self):
        """Test factor max limit constraint."""

        model = cvx.FactorMaxLimit(
            np.ones((self.N - 1, 2)), np.array([0.5, 1]))
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())

        model = cvx.FactorMaxLimit(np.ones((self.N - 1, 2)), np.array([4, 4]))
        cons = self._build_constraint(model, self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_factor_min_limit(self):
        """Test factor min limit constraint."""

        model = cvx.FactorMinLimit(
            np.ones((self.N - 1, 2)), np.array([0.5, 1]))
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3

        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        model = cvx.FactorMinLimit(np.ones((self.N - 1, 2)), np.array([4, 4]))
        cons = self._build_constraint(model, self.returns.index[1])
        # cons = model.weight_expr(t, self.w_plus, None, None)[0]
        tmp = np.zeros(self.N)
        tmp[0] = 4
        # tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_factor_gross_limit(self):
        """Test factor gross limit constraint."""

        model = cvx.FactorGrossLimit(
            np.ones((self.N - 1, 2)), np.array([0.5, 1]))
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())

        model = cvx.FactorGrossLimit(
            np.ones((self.N - 1, 2)), np.array([7, 7]))
        cons = self._build_constraint(model, self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_fixed_alpha(self):
        """Test fixed alpha constraint."""

        model = cvx.FixedFactorLoading(np.ones((self.N - 1, 1)), 1)
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_factor_neutral(self):
        """Test FactorNeutral constraint."""

        model = cvx.FactorNeutral(np.ones((self.N - 1, 1)))
        cons = self._build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -4
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_turnover_limit(self):
        """Test turnover limit constraint."""
        model = cvx.TurnoverLimit(0.1)
        cons = self._build_constraint(model)
        self.z.value = np.zeros(self.N)
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertTrue(cons.value())

        self.z.value[1] = 0.2
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertTrue(cons.value())

        self.z.value[2] = -0.01
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertFalse(cons.value())

    def test_participation_rate(self):
        """Test max participation rate constraints."""

        t = self.returns.index[1]

        # avg daily value limits.
        value = 1e6
        model = cvx.ParticipationRateLimit(
            self.volumes, max_fraction_of_volumes=0.1)
        model.initialize_estimator_recursive(
            universe=self.returns.columns,
            trading_calendar=self.returns.index)
        cons = model.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        model.values_in_time_recursive(t=t, current_portfolio_value=value)
        print(model.portfolio_value.value)
        # cons = model.weight_expr(t, None, z, value)[0]
        tmp = np.zeros(self.N)
        tmp[:-1] = self.volumes.loc[t].values / value * 0.05
        self.z.value = tmp
        self.assertTrue(cons.value())
        self.z.value = -100 * self.z.value  # -100*np.ones(n)
        self.assertFalse(cons.value())


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements the simple forecasting models used by Cvxportfolio.

These are standard ones like historical mean, variance, and covariance. In
most cases the models implemented here are equivalent to the relevant Pandas
DataFrame methods, including (most importantly) the logic used to skip over any
``np.nan``. There are some subtle differences explained below.

Our forecasters are optimized to be evaluated sequentially in time: at each
point in time in a back-test the forecast computed at the previous time step
is updated with the most recent observation. This is in some cases (e.g.,
covariances) much more efficient than computing from scratch.

Most of our forecasters implement both a rolling window and exponential moving
average logic. These are specified by the ``rolling`` and ``half_life``
parameters respectively, which are either Pandas Timedeltas or ``np.inf``.
The latter is the default, and means that the whole past is used, with no
exponential smoothing. Note that it's possible to use both, *e.g.*,
estimate covariance matrices ignoring past returns older than 5 years and
smoothing the recent ones using an exponential kernel with half-life of 1 year.

Finally, we note that the covariance, variance and standard deviation
forecasters implement the ``kelly`` parameter, which is True by default.
This is a simple trick explained in
:paper:`section 4.2 (page 28) <section.4.2>` of the paper, simplifies the
computation and provides in general (slightly) higher performance.
For example, using the notation of the paper, the classical definition of
covariance is

.. math::

    \Sigma = \mathbf{E}(r_t - \mu)(r_t - \mu)^T,

this is what you get by setting ``kelly=False``. The default, ``kelly=True``,
gives instead

.. math::

    \Sigma^\text{kelly} = \mathbf{E}r_t r_t^T = \Sigma + \mu \mu^T,

so that the resulting Markowitz-style optimization problem corresponds to
the second order Taylor approximation of a (risk-constrained) Kelly objective,
as is explained briefly :paper:`at page 28 of the paper <section.4.2>`, or with
more detail (and hard-to-read math) in `section 6 of the Risk-Constrained Kelly
Gambling paper
<https://web.stanford.edu/~boyd/papers/pdf/kelly.pdf#section.6>`_.

Lastly, some forecasters implement a basic caching mechanism.
This is used in two ways. First, online (e.g., in back-test): if multiple
copies of the same forecaster need access to the estimated value, as is the
case in :class:`cvxportfolio.MultiPeriodOptimization` policies, the expensive
evaluation is only done once. Then, offline, provided that the
:class:`cvxportfolio.data.MarketData` server used implements the
:meth:`cvxportfolio.data.MarketData.partial_universe_signature` method
(so that we can certify which market data the cached values are computed on).
This type of caching simply saves on disk the forecasted values, and makes it
available automatically next time the user runs a back-test on the same market
data (and same universe). This is especially useful when doing hyper-parameter
optimization, so that expensive computations like evaluating large covariance
matrices are only done once.

How to use them
~~~~~~~~~~~~~~~

These forecasters are each the default option of some Cvxportfolio optimization
term, for example :class:`HistoricalMeanReturn` is the default used by
:class:`cvxportfolio.ReturnsForecast`. In this way each is used with its
default options. If you want to change the options you can simply pass
the relevant forecaster class, instantiated with the options of your choice,
to the Cvxportfolio object. For example

.. code-block::

    import cvxportfolio as cvx
    from cvxportfolio.forecast import HistoricalMeanReturn
    import pandas as pd

    returns_forecast = cvx.ReturnsForecast(
        r_hat = HistoricalMeanReturn(
            half_life=pd.Timedelta(days=365),
            rolling=pd.Timedelta(days=365*5)))

if you want to apply exponential smoothing to the mean returns forecaster with
half-life of 1 year, and skip over all observations older than 5 years. Both
are relative to each point in time at which the policy is evaluated.
"""

import logging
from dataclasses import dataclass
from typing import Union

import numpy as np
import pandas as pd

from .errors import DataError, ForecastError
from .estimator import Estimator, SimulatorEstimator
from .hyperparameters import _resolve_hyperpar

logger = logging.getLogger(__name__)

def online_cache(values_in_time):
    """A simple online cache that decorates :meth:`values_in_time`.

    The instance it is used on needs to be hashable (we currently use
    the hash of its __repr__ via dataclass).

    :param values_in_time: :meth:`values_in_time` method to decorate.
    :type values_in_time: function

    :returns: Decorated method, which saves and retrieves (if available) from
        cache the result.
    :rtype: function
    """

    def wrapped(self, t, cache=None, **kwargs):
        """Cached :meth:`values_in_time` method.

        :param self: Class instance.
        :type self: cvxportfolio.Estimator
        :param t: Current time, used as in the key for caching.
        :type t: pandas.Timestamp
        :param cache: Cache dictionary; if None (the default) caching is
            disabled.
        :type cache: dict or None
        :param kwargs: Extra arguments that are passed through.
        :type kwargs: dict

        :returns: The returned value, maybe retrieved from cache.
        :rtype: float or numpy.array
        """

        if cache is None:  # temporary to not change tests
            cache = {}

        if not self in cache:
            cache[self] = {}

        if t in cache[self]:
            logger.debug(
                '%s.values_in_time at time %s is retrieved from cache.',
                self, t)
            result = cache[self][t]
        else:
            result = values_in_time(self, t=t, cache=cache, **kwargs)
            logger.debug('%s.values_in_time at time %s is stored in cache.',
                self, t)
            cache[self][t] = result
        return result

    return wrapped


class BaseForecast(Estimator):
    """Base class for forecasters."""

    _last_time = None

    def __post_init__(self):
        raise NotImplementedError # pragma: no cover

    def initialize_estimator( # pylint: disable=arguments-differ
            self, **kwargs):
        """Re-initialize whenever universe changes.

        :param kwargs: Unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self.__post_init__()

    def _agnostic_update(self, t, past_returns, **kwargs):
        """Choose whether to make forecast from scratch or update last one."""
        if (self._last_time is None) or (
            self._last_time != past_returns.index[-1]):
            logger.debug(
                '%s.values_in_time at time %s is computed from scratch.',
                self, t)
            self._initial_compute(t=t, past_returns=past_returns, **kwargs)
        else:
            logger.debug(
              '%s.values_in_time at time %s is updated from previous value.',
              self, t)
            self._online_update(t=t, past_returns=past_returns, **kwargs)

    def _initial_compute(self, **kwargs):
        """Make forecast from scratch."""
        raise NotImplementedError # pragma: no cover

    def _online_update(self, **kwargs):
        """Update forecast from period before."""
        raise NotImplementedError # pragma: no cover

def _is_timedelta(value):
    if isinstance(value, pd.Timedelta):
        if value <= pd.Timedelta('0d'):
            raise ValueError(
                '(Exponential) moving average window must be positive')
        return True
    if isinstance(value, float) and np.isposinf(value):
        return False
    raise ValueError(
        '(Exponential) moving average window can only be'
        ' pandas Timedeltas or np.inf.')


@dataclass(unsafe_hash=True)
class BaseMeanVarForecast(BaseForecast):
    """This class contains logic common to mean and (co)variance forecasters.

    It implements both moving average and exponential moving average, which
    can be used at the same time (e.g., ignore observations older than 5
    years and weight exponentially with half-life of 1 year the recent ones).

    Then, it implements the "online update" vs "compute from scratch" model,
    updating with a new observations is much cheaper than computing from
    scratch (especially for covariances).
    """

    half_life: Union[pd.Timedelta, float] = np.inf
    rolling: Union[pd.Timedelta, float] = np.inf

    def __post_init__(self):
        self._last_time = None
        self._denominator = None
        self._numerator = None

    def _compute_numerator(self, df, emw_weights):
        """Exponential moving window (optional) numerator."""
        raise NotImplementedError # pragma: no cover

    def _compute_denominator(self, df, emw_weights):
        """Exponential moving window (optional) denominator."""
        raise NotImplementedError # pragma: no cover

    def _update_numerator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied in this class.
        """
        raise NotImplementedError # pragma: no cover

    def _update_denominator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied in this class.
        """
        raise NotImplementedError # pragma: no cover

    def _dataframe_selector(self, **kwargs):
        """Return dataframe we work with.

        This method receives the **kwargs passed to :meth:`values_in_time`.
        """
        raise NotImplementedError # pragma: no cover

    def values_in_time( # pylint: disable=arguments-differ
            self, **kwargs):
        """Obtain current value of the historical mean of given dataframe.

        :param kwargs: All arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Historical means of given dataframe.
        :rtype: numpy.array
        """
        self._agnostic_update(**kwargs)
        return (self._numerator / self._denominator).values

    def _emw_weights(self, index, t):
        """Get weights to apply to the past obs for EMW."""
        index_in_halflifes = (index - t) / _resolve_hyperpar(self.half_life)
        return np.exp(index_in_halflifes * np.log(2))

    def _initial_compute(self, t, **kwargs): # pylint: disable=arguments-differ
        """Make forecast from scratch.

        This method receives the **kwargs passed to :meth:`values_in_time`.
        """
        df = self._dataframe_selector(t=t, **kwargs)

        # Moving average window logic
        if _is_timedelta(_resolve_hyperpar(self.rolling)):
            df = df.loc[df.index >= t-_resolve_hyperpar(self.rolling)]

        # If EMW, compute weights here
        if _is_timedelta(_resolve_hyperpar(self.half_life)):
            emw_weights = self._emw_weights(df.index, t)
        else:
            emw_weights = None

        self._denominator = self._compute_denominator(df, emw_weights)
        if np.min(self._denominator.values) == 0:
            raise ForecastError(
                f'{self.__class__.__name__} is given a dataframe with '
                    + 'at least a column that has no values.')
        self._numerator = self._compute_numerator(df, emw_weights)
        self._last_time = t

        # used by covariance forecaster
        return df, emw_weights

    def _online_update(self, t, **kwargs): # pylint: disable=arguments-differ
        """Update forecast from period before.

        This method receives the **kwargs passed to :meth:`values_in_time`.
        """
        df = self._dataframe_selector(t=t, **kwargs)
        last_row = df.iloc[-1]

        # if emw discount past
        if _is_timedelta(_resolve_hyperpar(self.half_life)):
            time_passed_in_halflifes = (
                self._last_time - t)/_resolve_hyperpar(self.half_life)
            discount_factor = np.exp(time_passed_in_halflifes * np.log(2))
            self._denominator *= discount_factor
            self._numerator *= discount_factor
        else:
            discount_factor = 1.

        # for emw we also need to discount last element
        self._denominator += self._update_denominator(
            last_row) * discount_factor
        self._numerator += self._update_numerator(last_row) * discount_factor

        # Moving average window logic: subtract elements that have gone out
        if _is_timedelta(_resolve_hyperpar(self.rolling)):
            observations_to_subtract, emw_weights_of_subtract = \
                self._remove_part_gone_out_of_ma(df, t)
        else:
            observations_to_subtract, emw_weights_of_subtract = None, None

        self._last_time = t

        # used by covariance forecaster
        return (
            discount_factor, observations_to_subtract, emw_weights_of_subtract)

    def _remove_part_gone_out_of_ma(self, df, t):
        """Subtract from numerator and denominator too old observations."""

        observations_to_subtract = df.loc[
            (df.index >= (self._last_time - _resolve_hyperpar(self.rolling)))
            & (df.index < (t - _resolve_hyperpar(self.rolling)))]

        # If EMW, compute weights here
        if _is_timedelta(_resolve_hyperpar(self.half_life)):
            emw_weights = self._emw_weights(observations_to_subtract.index, t)
        else:
            emw_weights = None

        self._denominator -= self._compute_denominator(
            observations_to_subtract, emw_weights)
        if np.min(self._denominator.values) == 0:
            raise ForecastError(
                f'{self.__class__.__name__} is given a dataframe with '
                + 'at least a column that has no values.')
        self._numerator -= self._compute_numerator(
            observations_to_subtract, emw_weights).fillna(0.)

        # used by covariance forecaster
        return observations_to_subtract, emw_weights


@dataclass(unsafe_hash=True)
class BaseMeanForecast(BaseMeanVarForecast): # pylint: disable=abstract-method
    """This class contains the logic common to the mean forecasters."""

    def _compute_numerator(self, df, emw_weights):
        """Exponential moving window (optional) numerator."""
        if emw_weights is None:
            return df.sum()
        return df.multiply(emw_weights, axis=0).sum()

    def _compute_denominator(self, df, emw_weights):
        """Exponential moving window (optional) denominator."""
        if emw_weights is None:
            return df.count()
        ones = (~df.isnull()) * 1.
        return ones.multiply(emw_weights, axis=0).sum()

    def _update_numerator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied upstream.
        """
        return last_row.fillna(0.)

    def _update_denominator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied upstream.
        """
        return ~(last_row.isnull())


@dataclass(unsafe_hash=True)
class HistoricalMeanReturn(BaseMeanForecast):
    r"""Historical means of non-cash returns.

    .. versionadded:: 1.2.0

        Added the ``half_life`` and ``rolling`` parameters.

    When both ``half_life`` and ``rolling`` are infinity, this is equivalent to

    .. code-block::

        past_returns.iloc[:,:-1].mean()

    where ``past_returns`` is a time-indexed dataframe containing the past
    returns (if in back-test that's relative to each point in time, ), and its
    last column, which we skip over, are the cash returns. We use the same
    logic as Pandas to handle ``np.nan`` values.

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. If in back-test, that is with respect to each point
        in time. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    """
    # pylint: disable=arguments-differ
    def _dataframe_selector(self, past_returns, **kwargs):
        """Return dataframe to compute the historical means of."""
        return past_returns.iloc[:, :-1]

@dataclass(unsafe_hash=True)
class HistoricalMeanVolume(BaseMeanForecast):
    r"""Historical means of traded volume in units of value (e.g., dollars).

    .. versionadded:: 1.2.0

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. If in back-test, that is with respect to each point
        in time. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    """
    # pylint: disable=arguments-differ
    def _dataframe_selector(self, past_volumes, **kwargs):
        """Return dataframe to compute the historical means of."""
        if past_volumes is None:
            raise DataError(
                f"{self.__class__.__name__} can only be used if MarketData"
                + " provides market volumes.")
        return past_volumes

@dataclass(unsafe_hash=True)
class HistoricalVariance(BaseMeanForecast):
    r"""Historical variances of non-cash returns.

    .. versionadded:: 1.2.0

        Added the ``half_life`` and ``rolling`` parameters.

    When both ``half_life`` and ``rolling`` are infinity, this is equivalent to

    .. code-block::

        past_returns.iloc[:,:-1].var(ddof=0)

    if you set ``kelly=False`` and

    .. code-block::

        (past_returns**2).iloc[:,:-1].mean()

    otherwise (we use the same logic to handle ``np.nan`` values).

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. If in back-test, that is with respect to each point
        in time. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    :param kelly: if ``True`` compute :math:`\mathbf{E}[r^2]`, else
        :math:`\mathbf{E}[r^2] - {\mathbf{E}[r]}^2`. The second corresponds
        to the classic definition of variance, while the first is what is
        obtained by Taylor approximation of the Kelly gambling objective.
        See discussion above.
    :type kelly: bool
    """

    kelly: bool = True

    def __post_init__(self):
        if not self.kelly:
            self.meanforecaster = HistoricalMeanReturn(
                half_life=_resolve_hyperpar(self.half_life),
                rolling=_resolve_hyperpar(self.rolling))
        super().__post_init__()

    def values_in_time(self, **kwargs):
        """Obtain current value either by update or from scratch.

        :param kwargs: All arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Variances of past returns (excluding cash).
        :rtype: numpy.array
        """
        result = super().values_in_time(**kwargs)
        if not self.kelly:
            result -= self.meanforecaster.current_value**2
        return result

    # pylint: disable=arguments-differ
    def _dataframe_selector(self, past_returns, **kwargs):
        """Return dataframe to compute the historical means of."""
        return past_returns.iloc[:, :-1]**2


@dataclass(unsafe_hash=True)
class HistoricalStandardDeviation(HistoricalVariance, SimulatorEstimator):
    """Historical standard deviation of non-cash returns.

    .. versionadded:: 1.2.0

        Added the ``half_life`` and ``rolling`` parameters.

    When both ``half_life`` and ``rolling`` are infinity, this is equivalent to

    .. code-block::

        past_returns.iloc[:,:-1].std(ddof=0)

    if you set ``kelly=False`` and

    .. code-block::

        np.sqrt((past_returns**2).iloc[:,:-1].mean())

    otherwise (we use the same logic to handle ``np.nan`` values).

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. If in back-test, that is with respect to each point
        in time. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    :param kelly: Same as in :class:`cvxportfolio.forecast.HistoricalVariance`.
        Default True.
    :type kelly: bool
    """

    kelly: bool = True

    def values_in_time(self, **kwargs):
        """Obtain current value either by update or from scratch.

        :param kwargs: All arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Standard deviations of past returns (excluding cash).
        :rtype: numpy.array
        """
        variances = \
            super().values_in_time(**kwargs)
        return np.sqrt(variances)

    def simulate(self, **kwargs):
        # TODO could take last return as well
        return self.values_in_time(
            t=kwargs['t'],
            # These are not necessary with current design of
            # DataEstimator
            current_weights=kwargs['current_weights'],
            current_portfolio_value=kwargs['current_portfolio_value'],
            past_returns=kwargs['past_returns'],
            past_volumes=kwargs['past_volumes'],
            current_prices=kwargs['current_prices']
        )

@dataclass(unsafe_hash=True)
class HistoricalMeanError(HistoricalVariance):
    r"""Historical standard deviations of the mean of non-cash returns.

    .. versionadded:: 1.2.0

        Added the ``half_life`` and ``rolling`` parameters.

    For a given time series of past returns :math:`r_{t-1}, r_{t-2},
    \ldots, r_0` this is :math:`\sqrt{\text{Var}[r]/t}`. When there are
    missing values we ignore them, both to compute the variance and the
    count.

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. If in back-test, that is with respect to each point
        in time. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    :param kelly: Same as in :class:`cvxportfolio.forecast.HistoricalVariance`.
        Default False.
    :type kelly: bool
    """

    kelly: bool = False

    def values_in_time(self, **kwargs):
        """Obtain current value either by update or from scratch.

        :param kwargs: All arguments to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Standard deviation of the mean of past returns (excluding
            cash).
        :rtype: numpy.array
        """
        variance = super().values_in_time(**kwargs)
        return np.sqrt(variance / self._denominator.values)


@dataclass(unsafe_hash=True)
class HistoricalCovariance(BaseMeanVarForecast):
    r"""Historical covariance matrix."""

    kelly: bool = True

    def __post_init__(self):
        super().__post_init__()
        self._joint_mean = None

    def _compute_numerator(self, df, emw_weights):
        """Exponential moving window (optional) numerator."""
        filled = df.fillna(0.)
        if emw_weights is None:
            return filled.T @ filled
        tmp = filled.multiply(emw_weights, axis=0)
        return tmp.T @ filled

    def _compute_denominator(self, df, emw_weights):
        """Exponential moving window (optional) denominator."""
        ones = (~df.isnull()) * 1.
        if emw_weights is None:
            return ones.T @ ones
        tmp = ones.multiply(emw_weights, axis=0)
        return tmp.T @ ones

    def _update_denominator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied upstream.
        """
        nonnull = ~(last_row.isnull())
        return np.outer(nonnull, nonnull)

    def _update_numerator(self, last_row):
        """Update with last observation.

        Emw (if any) is applied upstream.
        """
        filled = last_row.fillna(0.)
        return np.outer(filled, filled)

    def _dataframe_selector( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Return dataframe to compute the historical covariance of."""
        return past_returns.iloc[:, :-1]

    def _compute_joint_mean(self, df, emw_weights):
        r"""Compute precursor of :math:`\Sigma_{i,j} =
        \mathbf{E}[r^{i}]\mathbf{E}[r^{j}]`."""
        nonnull = (~df.isnull()) * 1.
        if emw_weights is None:
            return nonnull.T @ df.fillna(0.)
        return nonnull.T @ df.fillna(0.).multiply(emw_weights, axis=0)

    def _update_joint_mean(self, last_row):
        r"""Update precursor of :math:`\Sigma_{i,j} =
        \mathbf{E}[r^{i}]\mathbf{E}[r^{j}]`."""
        return last_row.fillna(0.)

    def _initial_compute( # pylint: disable=arguments-differ
            self, **kwargs):
        """Compute from scratch, taking care of non-Kelly correction."""

        df, emw_weights = super()._initial_compute(**kwargs)

        if not self.kelly:
            self._joint_mean = self._compute_joint_mean(df, emw_weights)

    def _online_update( # pylint: disable=arguments-differ
            self, **kwargs):
        """Update from last observation."""

        discount_factor, observations_to_subtract, emw_weights_of_subtract = \
            super()._online_update(**kwargs)
        df = self._dataframe_selector(**kwargs)
        last_row = df.iloc[-1]

        if not self.kelly:

            # discount past if EMW
            if discount_factor != 1.:
                self._joint_mean *= discount_factor

            # add last anyways
            self._joint_mean += self._update_joint_mean(
                last_row) * discount_factor

            # if MW, update by removing old observations
            if observations_to_subtract is not None:
                self._joint_mean -= self._compute_joint_mean(
                    observations_to_subtract, emw_weights_of_subtract)

    def values_in_time(self, **kwargs):
        """Obtain current value of the covariance estimate.

        :param kwargs: All arguments passed to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Covariance matrix (excludes cash).
        :rtype: numpy.array
        """

        covariance = super().values_in_time(**kwargs)

        if not self.kelly:
            tmp = self._joint_mean / self._denominator
            covariance -= tmp.T * tmp

        return covariance

def project_on_psd_cone_and_factorize(covariance):
    """Factorize matrix and remove negative eigenvalues.

    :param covariance: Square (symmetric) approximate covariance matrix, can
         have negative eigenvalues.
    :type covariance: numpy.array

    :returns: Square root factorization with negative eigenvalues removed.
    :rtype: numpy.array
    """
    eigval, eigvec = np.linalg.eigh(covariance)
    eigval = np.maximum(eigval, 0.)
    return eigvec @ np.diag(np.sqrt(eigval))

@dataclass(unsafe_hash=True)
class HistoricalFactorizedCovariance(HistoricalCovariance):
    r"""Historical covariance matrix of non-cash returns, factorized.

    .. versionadded:: 1.2.0

        Added the ``half_life`` and ``rolling`` parameters.

    When both ``half_life`` and ``rolling`` are infinity, this is equivalent
    to, before factorization

    .. code-block::

        past_returns.iloc[:,:-1].cov(ddof=0)

    if you set ``kelly=False``. We use the same logic to handle ``np.nan``
    values. For ``kelly=True`` it is not possible to reproduce with one single
    Pandas method (but we do test against Pandas in the unit tests).

    :param half_life: Half-life of exponential smoothing, expressed as
        Pandas Timedelta. Default ``np.inf``, meaning no exponential smoothing.
    :type half_life: pandas.Timedelta or np.inf
    :param rolling: Rolling window used: observations older than this Pandas
        Timedelta are skipped over. If in back-test, that is with respect to
        each point in time. Default ``np.inf``, meaning that all past is used.
    :type rolling: pandas.Timedelta or np.inf
    :param kelly: if ``True`` each element of the covariance matrix
        :math:`\Sigma_{i,j}` is equal to :math:`\mathbf{E} r^{i} r^{j}`,
        otherwise it is
        :math:`\mathbf{E} r^{i} r^{j} - \mathbf{E} r^{i} \mathbf{E} r^{j}`.
        The second case corresponds to the classic definition of covariance,
        while the first is what is obtained by Taylor approximation of
        the Kelly gambling objective. (See discussion above.)
        In the second case, the estimated covariance is the same
        as what is returned by ``pandas.DataFrame.cov(ddof=0)``, *i.e.*,
        we use the same logic to handle missing data.
    :type kelly: bool
    """

    # this is used by FullCovariance
    FACTORIZED = True

    @online_cache
    def values_in_time( # pylint: disable=arguments-differ
            self, t, **kwargs):
        """Obtain current value of the covariance estimate.

        :param t: Current time period (possibly of simulation).
        :type t: pandas.Timestamp
        :param kwargs: All arguments passed to :meth:`values_in_time`.
        :type kwargs: dict

        :raises cvxportfolio.errors.ForecastError: The procedure failed,
            typically because there are too many missing values (*e.g.*, some
            asset has only missing values).

        :returns: Square root factorized covariance matrix (excludes cash).
        :rtype: numpy.array
        """

        covariance = super().values_in_time(t=t, **kwargs)

        try:
            return project_on_psd_cone_and_factorize(covariance)
        except np.linalg.LinAlgError as exc: # pragma: no cover
            raise ForecastError(f'Covariance estimation at time {t} failed;'
                + ' there are (probably) too many missing values in the'
                + ' past returns.') from exc


@dataclass(unsafe_hash=True)
class HistoricalLowRankCovarianceSVD(Estimator):
    """Build factor model covariance using truncated SVD.

    .. note::

        This forecaster is experimental and not covered by semantic versioning,
        we may change it without warning.

    :param num_factors: How many factors in the low rank model.
    :type num_factors: int
    :param svd_iters: How many iteration of truncated SVD to apply. If you
        get a badly conditioned covariance you may to lower this.
    :type svd_iters: int
    :param svd: Which SVD routine to use, currently only dense (LAPACK) via
        Numpy.
    :type svd: str
    """

    num_factors: int
    svd_iters: int = 10
    svd: str = 'numpy'

    # brought back from old commit;
    #
    #  https://github.com/cvxgrp/cvxportfolio/commit
    #     /aa3d2150d12d85a6fb1befdf22cb7967fcc27f30
    #
    # matches original 2016 method from example
    # notebooks with new heuristic for NaNs. can probably be improved
    # by terminating early if idyosyncratic becomes negative

    @staticmethod
    def build_low_rank_model(rets, num_factors=10, iters=10, svd='numpy'):
        r"""Build a low rank risk model from past returns that include NaNs.

        This is an experimental procedure that may work well on past
        returns matrices with few NaN values (say, below 20% of the
        total entries). If there are (many) NaNs, one should probably
        also use a rather large risk forecast error.

        :param rets: Past returns, excluding cash.
        :type rets: pandas.DataFrame
        :param num_factors: How many factors in the fitted model.
        :type num_factors: int
        :param iters: How many iterations of SVD are performed.
        :type iters: int
        :param svd: Which singular value decomposition routine is used,
            default (and currently the only one supported) is ``'numpy'``.
        :type svd: str

        :raises SyntaxError: If wrong ``svd`` parameter is supplied.
        :raises cvxportfolio.errors.ForecastError: If the procedure fails;
             you may try with lower ``num_factors`` or ``iters``.

        :returns: (F, d)
        :rtype: (numpy.array, numpy.array)
        """
        nan_fraction = rets.isnull().sum().sum() / np.prod(rets.shape)
        normalizer = np.sqrt((rets**2).mean())
        normalized = rets
        if nan_fraction:
            nan_implicit_imputation = pd.DataFrame(0.,
                columns=normalized.columns, index = normalized.index)
            for _ in range(iters):
                if svd == 'numpy':
                    u, s, v = np.linalg.svd(
                        normalized.fillna(nan_implicit_imputation),
                        full_matrices=False)
                else:
                    raise SyntaxError(
                        'Currently only numpy svd is implemented')
                nan_implicit_imputation = pd.DataFrame(
                    (u[:, :num_factors] * (s[:num_factors]
                        )) @ v[:num_factors],
                    columns = normalized.columns, index = normalized.index)
        else:
            if svd == 'numpy':
                u, s, v = np.linalg.svd(normalized, full_matrices=False)
            else:
                raise SyntaxError(
                    'Currently only numpy svd is implemented')
        F = v[:num_factors].T * s[:num_factors] / np.sqrt(len(rets))
        F = pd.DataFrame(F.T, columns=normalizer.index)
        idyosyncratic = normalizer**2 - (F**2).sum(0)
        if not np.all(idyosyncratic >= 0.):
            raise ForecastError(
                "Low rank risk estimation with iterative SVD did not work."
                + " You probably have too many missing values in the past"
                + " returns. You may try with HistoricalFactorizedCovariance,"
                + " or change your universe.")
        return F.values, idyosyncratic.values

    @online_cache
    def values_in_time( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Current low-rank model, also cached.

        :param past_returns: Past market returns (including cash).
        :type past_returns: pandas.DataFrame
        :param kwargs: Extra arguments passed to :meth:`values_in_time`.
        :type kwargs: dict

        :returns: Low-rank plus diagonal covariance model: (F, d); excludes
             cash.
        :rtype: (numpy.array, numpy.array)
        """
        return self.build_low_rank_model(past_returns.iloc[:, :-1],
            num_factors=self.num_factors,
            iters=self.svd_iters, svd=self.svd)
